/*	Winnov Videum
 *	OSS Sound Driver
 *
 *	This program is copyright 1999
 *	by Bill Dirks
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Audio capture/playback and audio mixer module for Winnov Videum.
 *	This module depends on drivers/char/wnv.o.
 *
 *	Note: Only works with Videum cards having a WAVI-97 or later
 *	      Winnov chip, and a CS4218 sound chip. Which basically
 *	      means only Videum PCI AV, PnP-ISA or maybe some 
 *	      recent model legacy-ISA cards. No Traveler, nor most
 *	      non-PnP ISA cards. [1999-8-19]
 */


#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/sched.h>
#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/interrupt.h>
#include <linux/soundcard.h>
#include <linux/sound.h>

#define PKMOD "wnvaud: "
#include "../char/wnv.h"

#define AUDIO_DMA_BUFFER_SIZE		(1<<16)  /* bytes */

/*
 *	W A V I   H E L P E R   F U N C T I O N S
 */

static int
wavi_read_audio_pointer(struct videum_device *dev)
{
	int	i;
	__u16	a1, a2;

	for (i = 0; i < 10; ++i)
	{
		/* Read is unreliable, so read twice see if equal */
		a1 = wavi_readreg(dev, WAVI_PTRB) & 0xFF;
		a2 = wavi_readreg(dev, WAVI_PTRB) & 0xFF;
		if (a1 == a2)
			return ((int)a1 << 5);
	}
	return 0;
}

static int  /* returns previous value of ACE bit */
wavi_audio_converter_enable(struct videum_device *dev, int x)
{
	__u16	o = wavi_readreg(dev, WAVI_FREQ);
	__u16	t = o & WAVI_ACEBIT;
	wavi_writereg(dev, WAVI_FREQ, (x) ? t | WAVI_ACEBIT : t);
	return (o != 0);
}

static void
wavi_ams(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_FREQ) & ~WAVI_AMSBIT;
	wavi_writereg(dev, WAVI_FREQ, (x) ? t | WAVI_AMSBIT : t);
}

static void
wavi_freq(struct videum_device *dev, int freq)
{
	__u16	t = wavi_readreg(dev, WAVI_FREQ) & ~WAVI_FREQFIELD;
	wavi_writereg(dev, WAVI_FREQ, t | (freq & WAVI_FREQFIELD));
}

static void
wavi_imi(struct videum_device *dev, int imi)
{/*	0 = n * 8KHz   1 = n * 11KHz  */
	__u16	t = wavi_readreg(dev, WAVI_CTL) & ~WAVI_IMIBIT;
	t &= ~(WAVI_I2CBIT | WAVI_IMDBIT);
	t |= (dev->ctl & (WAVI_I2CBIT | WAVI_IMDBIT));
	dev->wavi_IMIBIT = (imi) ? WAVI_IMIBIT : 0;
	wavi_writereg(dev, WAVI_CTL, t | dev->wavi_IMIBIT);
}

static void
wavi_ass2(struct videum_device *dev, int src)
{/*	codec audio source select output pins  */
	/*	On most Videums:
	 *	0 = Line
	 *	1 = MXC audio
	 *	2 = Mic
	 *	3 = CD
	 */
	__u16	t = dev->wavi_ATTN & ~WAVI_ASS2FIELD;
	t |= (src << 2) & WAVI_ASS2FIELD;
	wavi_writereg(dev, WAVI_ATTN, t);
	dev->wavi_ATTN = t;
}

static void
wavi_islisr(struct videum_device *dev, int islisr)
{/*	codec input source select (left and right)  */
	/*	On most Videums this is always set to zero  */
	__u16	t = dev->wavi_GAIN & ~(WAVI_ISLBIT | WAVI_ISRBIT);
	t |= (islisr) ? 0 : (WAVI_ISLBIT | WAVI_ISRBIT);
	wavi_writereg(dev, WAVI_GAIN, t);
	dev->wavi_GAIN = t;
}

static void
wavi_outatt5(struct videum_device *dev, int left, int right)
{/*	4218 uses 5-bit output attenuation controls  */
	__u16	t = dev->wavi_ATTN & ~WAVI_4218_ATTFIELD;
	t |= (~left << 9) & WAVI_ATTLFFIELD;
	t |= (~right << 4) & WAVI_ATTRTFIELD;
	wavi_writereg(dev, WAVI_ATTN, t);
	dev->wavi_ATTN = t;
}

static void
wavi_ingain(struct videum_device *dev, int left, int right)
{/*	Each gain control is 0..15 (4 bits)  */
	__u16	t = dev->wavi_GAIN & ~WAVI_GAINFIELD;
	t |= (left << 4) & WAVI_GAINLFFIELD;
	t |= (right) & WAVI_GAINRTFIELD;
	wavi_writereg(dev, WAVI_GAIN, t);
	dev->wavi_GAIN = t;
}

static void
wavi_mixer(struct videum_device *dev, int mix, int v)
{/*	mix = 0..4; v = 0..255;  */
	/*	On most Videums
	 *	0 (MIX1L) = WAVE out volume   (MIXER_PCM)
	 *	1 (MIX1R) = CD monitor volume (MIXER_CD)
	 *	2 (MIX2L) = Left audio out    (MIXER_VOLUME left)
	 *	3 (MIX2R) = Right audio out   (MIXER_VOLUME right)
	 *	4 (AOVOL) = Record monitor    (MIXER_IMIX)
	 */
	__u16	t = 0;
	switch (mix)
	{
	case 0:
		t = dev->wavi_MIX1 & ~WAVI_MIXLFIELD;
		wavi_writereg(dev, WAVI_AUDMIX1, t | (v & WAVI_MIXLFIELD));
		dev->wavi_MIX1 = t | (v & WAVI_MIXLFIELD);
		break;
	case 1:
		t = dev->wavi_MIX1 & ~WAVI_MIXRFIELD;
		wavi_writereg(dev, WAVI_AUDMIX1, 
			      t | ((v << 8) & WAVI_MIXRFIELD));
		dev->wavi_MIX1 = t | ((v << 8) & WAVI_MIXRFIELD);
		break;
	case 2:
		t = dev->wavi_MIX2 & ~WAVI_MIXLFIELD;
		wavi_writereg(dev, WAVI_AUDMIX2, t | (v & WAVI_MIXLFIELD));
		dev->wavi_MIX2 = t | (v & WAVI_MIXLFIELD);
		break;
	case 3:
		t = dev->wavi_MIX2 & ~WAVI_MIXRFIELD;
		wavi_writereg(dev, WAVI_AUDMIX2, 
			      t | ((v << 8) & WAVI_MIXRFIELD));
		dev->wavi_MIX2 = t | ((v << 8) & WAVI_MIXRFIELD);
		break;
	case 4:
		t = wavi_readreg(dev, WAVI_FREQ) & ~WAVI_AOVFIELD;
		wavi_writereg(dev, WAVI_FREQ, t | ((v << 8) & WAVI_AOVFIELD));
		break;
	}
}

static int
wavi_initialize_audio(struct videum_device *dev)
{
	if (dev->eeprom.dwHwFlags & HWFLAGS_CS4218)
	{
		wavi_writereg(dev, WAVI_MMA, wavi_readreg(dev, WAVI_MMA)
			      | WAVI_CS4218BIT);
		wavi_ams(dev, 1);
	}
	else
	{
		err_msg("%s- No CS4218 audio chip\n", dev->shortname);
		return 0;
	}
	if (dev->eeprom.dwHwFlags & HWFLAGS_AUDXSEL)
	{
		wavi_writereg(dev, WAVI_MMA, wavi_readreg(dev, WAVI_MMA)
			      | WAVI_AUDXSEL_INT97);
	}
	return 1;
}


/*
 *
 *	M I X E R   F U N C T I O N S
 *
 */

static int
mixer_pciav_source(struct videum_device *dev,
		   int	src)
{
	int	as;

	if (src == 0) /*  The OSS spec says zero = Mic  */
		src = SOUND_MASK_MIC;
	if (src == dev->mix_source) /*  No change  */
		return 1;
	/*  Can only select one source, so switch off old source  */
	if (src & dev->mix_source)
		src ^= dev->mix_source;
	/*  Make sure only one bit is set  */
	if (src & SOUND_MASK_LINE)  src = SOUND_MASK_LINE;
	if (src & SOUND_MASK_MIC)   src = SOUND_MASK_MIC;
	if (src & SOUND_MASK_CD)    src = SOUND_MASK_CD;
	if (src & SOUND_MASK_VIDEO) src = SOUND_MASK_VIDEO;
	switch(src)
	{
	case SOUND_MASK_LINE:	as = 0;  break;
	case SOUND_MASK_MIC:	as = 2;  break;
	case SOUND_MASK_CD:	as = 3;  break;
	case SOUND_MASK_VIDEO:	as = 1;  break;
	default:		return 0;
	}
	wavi_ass2(dev, as);
	dev->mix_source = src;
	return 1;
}

static int
mixer_pciav_init(struct videum_device *dev)
{
	/*  Which channels have programmable level controls */
	dev->mix_devmask =
		SOUND_MASK_RECLEV  |  /* Recording input gain */
		SOUND_MASK_IMIX    |  /* Record source monitor */
		SOUND_MASK_CD      |  /* CD monitor volume  */
		SOUND_MASK_PCM     |  /* WAVE out volume  */
		SOUND_MASK_VOLUME  |  /* Global output volume  */
		0;
	/*  Which level controls are stereo  */
	dev->mix_stereodevs =
		SOUND_MASK_RECLEV  |  /* Record input  */
		SOUND_MASK_VOLUME  |  /* Output volume  */
		0;
	/*  Recording inputs  */
	dev->mix_recmask =
		SOUND_MASK_LINE    |  /* Line  */
		SOUND_MASK_MIC     |  /* Mic  */
		SOUND_MASK_CD 	   |  /* CD   */
		SOUND_MASK_VIDEO   |  /* MXC camera/TV audio  */
		0;

	dev->mix_source	= 0;
	dev->mix_reclev	= 50 * 256 + 50;
	dev->mix_imix	= 50 * 256 + 50;
	dev->mix_cd	= 50 * 256 + 50;
	dev->mix_pcm	= 50 * 256 + 50;
	dev->mix_volume	= 50 * 256 + 50;
	wavi_islisr(dev, 0);
	mixer_pciav_source(dev, SOUND_MASK_LINE);
	wavi_ingain(dev, 8, 8);
	wavi_outatt5(dev, 31, 31);
	wavi_mixer(dev, 0, 128); /* WAVE out */
	wavi_mixer(dev, 1, 128); /* CD */
	wavi_mixer(dev, 2, 128); /* left volume */
	wavi_mixer(dev, 3, 128); /* right volume */
	wavi_mixer(dev, 4, 128); /* monitor */
	return 1;
}

static void
mixer_shutdown(struct videum_device *dev)
{
	wavi_mixer(dev, 0, 0);
	wavi_mixer(dev, 1, 0);
	wavi_mixer(dev, 2, 0);
	wavi_mixer(dev, 3, 0);
	wavi_mixer(dev, 4, 0);
}

static int
scale_100_16(int x)
{
	x = 15 * x / 100;
	if (x < 0) x = 0;
	if (x > 15) x = 15;
	return x;
}
static int
scale_100_256(int x)
{
	x = 255 * x / 100;
	if (x < 0) x = 0;
	if (x > 255) x = 255;
	return x;
}


/*
 *	MIXER API INTERFACE
 */

static int
mixer_open(struct inode *inode, struct file *file)
{
	int			minor	= MINOR(inode->i_rdev);
	int			i;
	struct videum_device	*dev	= NULL;

	for (i = 0; i < WNV_NBOARDS; ++i)
		if (videum[i].mixer_unit == minor)
		{
			dev = &videum[i];
			break;
		}
	if (dev == NULL)
		return -ENODEV;
	file->private_data = dev;

	//MOD_INC_USE_COUNT;
	//debug_msg("Opened %s mixer\n",dev->shortname);
	return 0;
}

static int
mixer_close(struct inode *inode, struct file *file)
{
  //	struct videum_device *dev = file->private_data;

	//MOD_DEC_USE_COUNT;
	//debug_msg("Closed %s mixer\n",dev->shortname);
	return 0;
}

static int
mixer_ioctl(struct inode *inode, struct file *file,
	    unsigned int cmd, unsigned long arg)
{
	struct videum_device	*dev	= file->private_data;
	int			lf, rt;
	int			val;

	switch (cmd)
	{
	case OSS_GETVERSION:
	{
		return put_user(SOUND_VERSION, (int *)arg);
	}
	case SOUND_MIXER_INFO:
	{
		mixer_info	minfo;
		strncpy(minfo.id, dev->shortname, sizeof(minfo.id));
		strncpy(minfo.name, dev->v.name, sizeof(minfo.name));
		minfo.modify_counter = 0;
		if (copy_to_user((void *)arg, &minfo, sizeof(minfo)))
			return -EFAULT;
		return 0;
	}

	case SOUND_MIXER_READ_DEVMASK:
	{
		return put_user(dev->mix_devmask, (int *)arg);
	}
	case SOUND_MIXER_READ_RECMASK:
	{
		return put_user(dev->mix_recmask, (int *)arg);
	}
	case SOUND_MIXER_READ_STEREODEVS:
	{
		return put_user(dev->mix_stereodevs, (int *)arg);
	}
	case SOUND_MIXER_READ_CAPS:
	{
		return put_user(SOUND_CAP_EXCL_INPUT, (int *)arg);
	}

	case MIXER_READ(SOUND_MIXER_RECLEV):
	{
		return put_user(dev->mix_reclev, (int *)arg);
	}
	case MIXER_READ(SOUND_MIXER_IMIX):
	{
		return put_user(dev->mix_imix, (int *)arg);
	}
	case MIXER_READ(SOUND_MIXER_CD):
	{
		return put_user(dev->mix_cd, (int *)arg);
	}
	case MIXER_READ(SOUND_MIXER_PCM):
	{
		return put_user(dev->mix_pcm, (int *)arg);
	}
	case MIXER_READ(SOUND_MIXER_VOLUME):
	{
		return put_user(dev->mix_volume, (int *)arg);
	}

	case MIXER_WRITE(SOUND_MIXER_RECLEV):
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		lf = scale_100_16(val & 0xff);
		rt = scale_100_16((val & 0xff00) >> 8);
		wavi_ingain(dev, lf, rt);
		dev->mix_reclev = val & 0xffff;
		return put_user(dev->mix_reclev, (int *)arg);
	}
	case MIXER_WRITE(SOUND_MIXER_IMIX):
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		lf = scale_100_256(val &= 0xff);
		wavi_mixer(dev, 4, lf);
		dev->mix_imix = (val << 8) | val;
		return put_user(dev->mix_imix, (int *)arg);
	}
	case MIXER_WRITE(SOUND_MIXER_CD):
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		lf = scale_100_256(val &= 0xff);
		wavi_mixer(dev, 1, lf);
		dev->mix_cd = (val << 8) | val;
		return put_user(dev->mix_cd, (int *)arg);
	}
	case MIXER_WRITE(SOUND_MIXER_PCM):
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		lf = scale_100_256(val &= 0xff);
		wavi_mixer(dev, 0, lf);
		dev->mix_pcm = (val << 8) | val;
		return put_user(dev->mix_pcm, (int *)arg);
	}
	case MIXER_WRITE(SOUND_MIXER_VOLUME):
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		lf = scale_100_256(val & 0xff);
		rt = scale_100_256((val & 0xff00) >> 8);
		wavi_mixer(dev, 2, lf);
		wavi_mixer(dev, 3, rt);
		dev->mix_volume = val & 0xffff;
		return put_user(dev->mix_volume, (int *)arg);
	}

	case SOUND_MIXER_READ_RECSRC:
	{
		return put_user(dev->mix_source, (int *)arg);
	}
	case SOUND_MIXER_WRITE_RECSRC:
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		if (!mixer_pciav_source(dev, val))
			return -EINVAL;
		return put_user(dev->mix_source, (int *)arg);
	}

	}
debug_msg("Unhandled mixer ioctl %08X\n",cmd);
	return -ENOIOCTLCMD;
}

static loff_t
mixer_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static struct file_operations mixer_fops =
{
	mixer_llseek,
	NULL,  /* read */
	NULL,  /* write */
	NULL,  /* readdir */
	NULL,  /* poll */
	mixer_ioctl,
	NULL,  /* mmap */
	mixer_open,
	NULL,  /* flush */
	mixer_close,
	NULL,  /* fsync */
	NULL,  /* fasync */
	NULL,  /* check_media_change */
	NULL,  /* revalidate */
	NULL,  /* lock */
};


/*
 *
 *	A U D I O   F U N C T I O N S
 *	(WAVE)
 */

static void
audio_fragment_size(struct videum_device *dev)
{	/*  Recompute fragment size  */
	if (dev->a_dma_enabled)
		return;
	dev->a_fragment = 4096;
	if (dev->a_speed <= 16000)
		dev->a_fragment = 2048;
}

static void
audio_read_one_fragment(struct videum_device *dev,
			int	count)
{
	wavi_block_read(dev, dev->sram_audio
			 + dev->a_record_audptr,
			 dev->a_record_xfer_ptr,
			 count);

	/*  Advance pointers:  */
	/*  This is the record "DMA" buffer pointer  */
	dev->a_record_xfer_ptr += count;
	if (dev->a_record_xfer_ptr - dev->a_record_buffer >=
	    AUDIO_DMA_BUFFER_SIZE)
		dev->a_record_xfer_ptr = dev->a_record_buffer;
	/*  This tracks recorded data in "DMA" buffer  */
	dev->a_record_data_avail += count;

	/*  This tracks where the next read from SRAM should come from  */
	dev->a_record_audptr += count;
	if (dev->a_record_audptr >= WAVI_AUDBUFSIZE)
		dev->a_record_audptr = 0;
	/*  This tracks recorded data in the SRAM audio buffer  */
	dev->a_record_sram_data -= count;

	/*  Look for overrun condition (if the app stops reading from	*/
	/*  the DMA buffer) and fixup the variables			*/
	if (dev->a_record_data_avail > AUDIO_DMA_BUFFER_SIZE)
	{
		debug_msg("Record overrun\n");
		dev->a_record_data_avail = AUDIO_DMA_BUFFER_SIZE;
		dev->a_record_read_ptr = dev->a_record_xfer_ptr;
	}
}

static void
audio_write_one_fragment(struct videum_device *dev,
			 int	count)
{
	wavi_block_write(dev,
			 dev->sram_audio + dev->a_play_audptr,
			 dev->a_play_xfer_ptr,
			 count, 0);
	/*  Keep zeros in unused space in buffer  */
	memset(dev->a_play_xfer_ptr, 0, count);

	/*  Advance pointers:  */
	/*  This is the play "DMA" buffer pointer  */
	dev->a_play_xfer_ptr += count;
	if (dev->a_play_xfer_ptr - dev->a_play_buffer >=
	    AUDIO_DMA_BUFFER_SIZE)
		dev->a_play_xfer_ptr = dev->a_play_buffer;
	/*  This tracks space available in "DMA" buffer  */
	dev->a_play_space_avail += count;

	/*  This tracks where the next write to SRAM should go  */
	dev->a_play_audptr += count;
	if (dev->a_play_audptr >= WAVI_AUDBUFSIZE)
		dev->a_play_audptr = 0;
	/*  This tracks space available in the SRAM audio buffer  */
	dev->a_play_sram_avail -= count;

	/*  Look for underrun condition (if the app stops writing to  */
	/*  the DMA buffer) and fixup the variables		      */
	if (dev->a_play_space_avail > AUDIO_DMA_BUFFER_SIZE)
	{
		debug_msg("Play underrun\n");
		dev->a_play_space_avail = AUDIO_DMA_BUFFER_SIZE;
		dev->a_play_write_ptr = dev->a_play_xfer_ptr;
	}
}

/*
 *	V I R T U A L   D M A   B U F F E R S
 */

static void
vdma_poll(void *v)
{
	struct videum_device	*dev = (struct videum_device *)v;
	int			audptr;
	int			delta;
	int			tocopy	= 0;

	if (dev == NULL || !dev->a_dma_enabled)
		return;

	/*  New audio pointer and change since last poll  */
	audptr = wavi_read_audio_pointer(dev);
	delta = audptr - dev->a_last_audptr;
	if (delta < 0)
		delta += WAVI_AUDBUFSIZE;
	//if (delta == 0) debug_msg("AUDPTR didn't move!\n");

	/*  Get new record data into record buffer  */
	if (dev->a_do_record_xfers)
	{
		/*  Data in SRAM audio buffer  */
		dev->a_record_sram_data += delta;
		if (dev->a_record_sram_data < dev->a_fragment)
			tocopy = 0;
		else
			tocopy = dev->a_fragment;

		//debug_msg("AP: %d  recAP: %d  data: %d  Get %d bytes\n",
		//audptr,dev->a_record_audptr,dev->a_record_sram_data,tocopy);
		if (tocopy)
		{
			audio_read_one_fragment(dev, tocopy);
			/*  Announce new data available in the buffer  */
			wake_up_interruptible(&dev->a_record_wait);
		}
	}

	/*  Send next fragment of play data  */
	if (dev->a_do_play_xfers)
	{
		/*  Room in SRAM audio buffer  */
		dev->a_play_sram_avail += delta;
		if (dev->a_play_sram_avail < dev->a_fragment)
			tocopy = 0;
		else
			tocopy = dev->a_fragment;

		//debug_msg("AP: %d  playAP: %d  avail: %d  Play %d bytes\n",
		//audptr,dev->a_play_audptr,dev->a_play_sram_avail,tocopy);
		if (tocopy)
		{
			audio_write_one_fragment(dev, tocopy);
			/*  Announce new space available in the buffer  */
			wake_up_interruptible(&dev->a_play_wait);
		}
	}

	/*  Keep tallies  */
	if (tocopy)
		++dev->a_total_fragments;
	dev->a_total_bytes += delta;
	dev->a_last_audptr = audptr;
}

static void
vdma_interrupt(unsigned long v)
{
	struct videum_device *dev = (struct videum_device *)v;

	if (!dev->a_dma_enabled)
	{
		return;
	}

#if 1
	/*  Call "bottom half" of handler  */
	dev->a_tqnode_dpc.routine = vdma_poll;
	dev->a_tqnode_dpc.data = dev;
	queue_task(&dev->a_tqnode_dpc, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
#else
	vdma_poll((void *)v);
#endif

	/*  Poll again on next timer tick  */
	dev->a_tlnode.expires = jiffies + HZ/100;
	add_timer(&dev->a_tlnode);
}

static int
vdma_run(struct videum_device *dev)
{
	if (dev->a_record_buffer == NULL &&
	    dev->a_play_buffer == NULL)
		return 0;
	if (dev->a_dma_enabled)
		return 1;

	/*  Compute fragment size  */
	audio_fragment_size(dev);

	/*  Turn on WAVI audio  */
	wavi_audio_converter_enable(dev, 0);
	wavi_freq(dev, dev->a_freq);
	wavi_imi(dev, dev->a_xtal);
	wavi_audio_converter_enable(dev, 1);
	wavi_ams(dev, 1);/* Must be after audio converter enabled  */

	dev->a_dma_enabled = 1;
	init_timer(&dev->a_tlnode);
	dev->a_tlnode.function = vdma_interrupt;
	dev->a_tlnode.data = (unsigned long)dev;
	dev->a_tlnode.expires = jiffies + HZ/100;
	add_timer(&dev->a_tlnode);
//debug_msg("Audio virtual DMA interrupts enabled\n");
	return 1;
}

static void
vdma_stop(struct videum_device *dev)
{
	if (dev->a_dma_enabled)
	{
		dev->a_dma_enabled = 0;
		del_timer(&dev->a_tlnode);
	}
	wavi_audio_converter_enable(dev, 0);
	wavi_writereg(dev, WAVI_AUDPTR, dev->sram_audio);
	dev->a_last_audptr = 0;
	dev->a_record_read_ptr = dev->a_record_buffer;
	dev->a_record_xfer_ptr = dev->a_record_buffer;
	dev->a_record_data_avail = 0;
	dev->a_record_audptr = 0;
	dev->a_record_sram_data = 0;

	dev->a_play_write_ptr = dev->a_play_buffer;
	dev->a_play_xfer_ptr = dev->a_play_buffer;
	dev->a_play_space_avail = AUDIO_DMA_BUFFER_SIZE;
	dev->a_play_audptr = 0;
	dev->a_play_sram_avail = WAVI_AUDBUFSIZE;
	dev->a_total_bytes = 0;
	dev->a_total_fragments = 0;

	if (dev->a_play_buffer)
		memset(dev->a_play_buffer, 0, dev->a_play_space_avail);
//debug_msg("Audio virtual DMA interrupts disabled\n");
}

static void
vdma_shutdown(struct videum_device *dev)
{
	vdma_stop(dev);
	if (dev->a_record_buffer)
		vfree(dev->a_record_buffer);
	dev->a_record_buffer = NULL;
	if (dev->a_play_buffer)
		vfree(dev->a_play_buffer);
	dev->a_play_buffer = NULL;
}

static int
vdma_setup(struct videum_device *dev)
{
	vdma_shutdown(dev);
	
	dev->a_record_buffer = vmalloc(AUDIO_DMA_BUFFER_SIZE);
	if (dev->a_record_buffer == NULL)
		return 0;
	dev->a_play_buffer = vmalloc(AUDIO_DMA_BUFFER_SIZE);
	if (dev->a_play_buffer == NULL)
	{
		vfree(dev->a_record_buffer);
		dev->a_record_buffer = NULL;
		return 0;
	}

	vdma_stop(dev);  /*  initializes most variables  */
	dev->a_dma_enabled = 0;
	dev->a_do_record_xfers = 0;
	dev->a_do_play_xfers = 0;
	
	return 1;
}


static void
audio_drain_play_buffer(struct videum_device *dev)
{
	while (dev->a_dma_enabled && dev->a_do_play_xfers)
	{
		cli();
		if (dev->a_play_space_avail < AUDIO_DMA_BUFFER_SIZE)
			interruptible_sleep_on(&dev->a_play_wait);
		sti();
		if (dev->a_play_space_avail >= AUDIO_DMA_BUFFER_SIZE)
			return;
		if (signal_pending(current))
			return;
	}
}

static int
audio_set_speed(struct videum_device *dev, int speed)
{
	int	freq	= WAVI_FREQFIELD_07350;
	int	xtal	= 0;
	int	have8k	=!(dev->eeprom.dwHwFlags&HWFLAGS_XTAL_8K_NOT_PRESENT);
	int	have11k	=!(dev->eeprom.dwHwFlags&HWFLAGS_XTAL_11K_NOT_PRESENT);

	if      (speed <=  8000 && have8k)
	{
		xtal = 0; freq = WAVI_FREQFIELD_07350; speed =  8000;
	}
	else if (speed <= 11025 && have11k)
	{
		xtal = 1; freq = WAVI_FREQFIELD_11025; speed = 11025;
	}
	else if (speed <= 12000 && have8k)
	{
		xtal = 0; freq = WAVI_FREQFIELD_11025; speed = 12000;
	}
	else if (speed <= 16000 && have8k)
	{
		xtal = 0; freq = WAVI_FREQFIELD_14700; speed = 16000;
	}
	else if (speed <= 22050 && have11k)
	{	
		xtal = 1; freq = WAVI_FREQFIELD_22050; speed = 22050;
	}
	else if (speed <= 24000 && have8k)
	{
		xtal = 0; freq = WAVI_FREQFIELD_22050; speed = 24000;
	}
	else if (speed <= 32000 && have8k)
	{
		xtal = 0; freq = WAVI_FREQFIELD_29400; speed = 32000;
	}
	else if (speed <= 44100 || !have8k)
	{
		xtal = 1; freq = WAVI_FREQFIELD_44100; speed = 44100;
	}
	else
	{
		xtal = 0; freq = WAVI_FREQFIELD_44100; speed = 48000;
	}

	dev->a_xtal = xtal;
	dev->a_freq = freq;
	dev->a_speed = speed;
	return speed;
}


/*
 *	AUDIO API INTERFACE
 */

static int
audio_open(struct inode *inode, struct file *file)
{
	int			minor	= MINOR(inode->i_rdev);
	int			i;
	struct videum_device	*dev	= NULL;

	for (i = 0; i < WNV_NBOARDS; ++i)
		if (((videum[i].dsp_unit ^ minor) & ~0xf) == 0)
			dev = &videum[i];
	if (dev == NULL)
		return -ENODEV;
	file->private_data = dev;

	if (dev->a_open)
	{
		debug_msg("Device %s is busy\n",dev->shortname);
		return -EBUSY;
	}
	dev->a_open = 1;
	dev->a_open_mode = file->f_mode;

	//MOD_INC_USE_COUNT;
debug_msg("Opened %s dsp\n",dev->shortname);
	return 0;
}

static int
audio_close(struct inode *inode, struct file *file)
{
	struct videum_device *dev = file->private_data;

	if (file->f_mode & FMODE_WRITE && !(file->f_flags & O_NONBLOCK))
		audio_drain_play_buffer(dev);
	vdma_stop(dev);
	dev->a_do_record_xfers = 0;
	dev->a_do_play_xfers = 0;

	dev->a_open = 0;
	dev->a_open_mode = 0;

	//MOD_DEC_USE_COUNT;
debug_msg("Closed %s dsp\n",dev->shortname);
	return 0;
}

static ssize_t
audio_read(struct file *file,
	   char *buf, size_t count, loff_t *ppos)
{
	int			block	= !(file->f_flags & O_NONBLOCK);
	struct videum_device	*dev	= file->private_data;
	int			read	= 0;
	int			toread;
	__u8			*a_record_read_ptr;
	int			a_record_data_avail;

	if (!access_ok(VERIFY_WRITE, buf, count))
		return -EFAULT;

	/*  Loop until the entire amount of data requested is read  */
	while (count > 0)
	{
		if (!dev->a_dma_enabled)
		{/*  Start recording immediately upon first read  */
			dev->a_do_record_xfers = 1;
			vdma_run(dev);
		}

		/*  If no data is available, then either return or wait  */
		if (dev->a_record_data_avail == 0 && !block)
			return read;
		cli();
		if (dev->a_record_data_avail == 0)
			interruptible_sleep_on(&dev->a_record_wait);
		sti();

		/*  Read to end of virtual DMA buffer or all available	*/
		/*  data or fill remaining count, whichever is less	*/
		toread = count;

		/*  Interrupt routine can touch a_record_read_ptr and	*/
		/*  a_record_data_avail (but only in an overrun cond.)	*/
		cli();
		a_record_read_ptr = dev->a_record_read_ptr;
		a_record_data_avail = dev->a_record_data_avail;
		sti();
		if (toread > a_record_data_avail)
			toread = a_record_data_avail;
		if (toread > dev->a_record_buffer + AUDIO_DMA_BUFFER_SIZE
		    - a_record_read_ptr)
			toread = dev->a_record_buffer
				+ AUDIO_DMA_BUFFER_SIZE
				- a_record_read_ptr;
		copy_to_user(buf, a_record_read_ptr, toread);

		/*  Advance pointers and account for read data  */
		count -= toread;
		buf += toread;
		read += toread;
		a_record_data_avail -= toread;
		a_record_read_ptr += toread;
		if (a_record_read_ptr >= dev->a_record_buffer
		    + AUDIO_DMA_BUFFER_SIZE)
			a_record_read_ptr = dev->a_record_buffer;
		cli();
		dev->a_record_read_ptr = a_record_read_ptr;
		dev->a_record_data_avail = a_record_data_avail;
		sti();
	}

	return read;
}

static ssize_t
audio_write(struct file *file,
	    const char *buf, size_t count, loff_t *ppos)
{
	int			block	= !(file->f_flags & O_NONBLOCK);
	struct videum_device	*dev	= file->private_data;
	ssize_t			written	= 0;
	int			towrite;
	__u8			*a_play_write_ptr;
	int			a_play_space_avail;

	if (!access_ok(VERIFY_READ, buf, count))
		return -EFAULT;

	/*  Loop until all data is written  */
	while (count > 0)
	{
		/*  If no space available in DMA buffer, exit or sleep  */
		if (dev->a_play_space_avail == 0 && !block)
			return written;
		cli();
		if (dev->a_play_space_avail == 0)/*  No room at the inn...  */
			interruptible_sleep_on(&dev->a_play_wait);
		sti();

		/*  Write to end of virtual DMA buffer or all remaining  */
		/*  data or fill remaining space, whichever is less	 */
		towrite = count;

		/*  Interrupt routine can touch a_play_write_ptr and	*/
		/*  a_play_space_avail (but only in an underrun cond.)	*/
		cli();
		a_play_write_ptr = dev->a_play_write_ptr;
		a_play_space_avail = dev->a_play_space_avail;
		sti();
		if (towrite > a_play_space_avail)
			towrite = a_play_space_avail;
		if (towrite > dev->a_play_buffer + AUDIO_DMA_BUFFER_SIZE
		    - a_play_write_ptr)
			towrite = dev->a_play_buffer
				+ AUDIO_DMA_BUFFER_SIZE
				- a_play_write_ptr;
		copy_from_user(a_play_write_ptr, buf, towrite);

		/*  Advance pointers and account for written data  */
		count -= towrite;
		buf += towrite;
		written += towrite;
		a_play_space_avail -= towrite;
		a_play_write_ptr += towrite;
		if (a_play_write_ptr >= dev->a_play_buffer
		    + AUDIO_DMA_BUFFER_SIZE)
			a_play_write_ptr = dev->a_play_buffer;
		cli();
		dev->a_play_write_ptr = a_play_write_ptr;
		dev->a_play_space_avail = a_play_space_avail;
		sti();

		if (!dev->a_dma_enabled && 
		    dev->a_play_space_avail <= 
		    AUDIO_DMA_BUFFER_SIZE - dev->a_fragment)
		{/*  Once a full fragment is written, start "DMA"  */
			audio_write_one_fragment(dev, dev->a_fragment);
			dev->a_do_play_xfers = 1;
			vdma_run(dev);
		}
	}
	return written;
}

static unsigned int
audio_poll(struct file *file, poll_table *table)
{
	struct videum_device	*dev	= file->private_data;
	unsigned int		ret	= 0;

	if (!dev->a_dma_enabled)  /*  Device is not running  */
		return POLLERR;

	if (dev->a_do_play_xfers && dev->a_play_space_avail)
	{
		if (file->f_mode & FMODE_WRITE)
			ret |= (POLLOUT | POLLWRNORM);
	}
	if (dev->a_do_record_xfers && 
	    dev->a_record_data_avail >= dev->a_fragment)
	{
		if (file->f_mode & FMODE_READ)
			ret |= (POLLIN | POLLRDNORM);
	}
	if (ret)
		return ret;

	/*  If device is open for reading, but the "DMA" isn't started	*/
	/*  then turn it on.   */
	if ((file->f_mode & FMODE_READ) && !dev->a_dma_enabled)
	{
		dev->a_do_record_xfers = 1;
		vdma_run(dev);
	}

	if (file->f_mode & FMODE_READ)
		poll_wait(file, &dev->a_record_wait, table);
	if (file->f_mode & FMODE_WRITE)
		poll_wait(file, &dev->a_play_wait, table);
	return 0;
}

static int
audio_ioctl(struct inode *inode, struct file *file,
	    unsigned int cmd, unsigned long arg)
{
	struct videum_device	*dev	= file->private_data;
	int			val;

	switch (cmd)
	{
	case OSS_GETVERSION:
	{
		return put_user(SOUND_VERSION, (int *)arg);
	}
	case SNDCTL_DSP_GETCAPS:
	{
		return put_user(0		|  /* GETCAPS version  */
				DSP_CAP_REVISION |
				//DSP_CAP_DUPLEX   | /* BUG: add me  */
				//DSP_CAP_REALTIME | /* BUG: add me  */
				//DSP_CAP_TRIGGER  | /* BUG: add me  */
				0, (int *)arg);
	}

	case SNDCTL_DSP_GETFMTS:
	{
		return put_user(AFMT_S16_LE	| /* signed 16b little-end */
				0, (int *)arg);
	}
	case SNDCTL_DSP_SETFMT:
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		if (val == AFMT_QUERY)
			return put_user(dev->a_format, (int *)arg);
		/*  We only support one format, so this is easy  */
		dev->a_format = AFMT_S16_LE;
		return put_user(dev->a_format, (int *)arg);
	}
	case SNDCTL_DSP_STEREO:
	case SNDCTL_DSP_CHANNELS:
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		if (cmd == SNDCTL_DSP_CHANNELS)
			--val;
		if (val == 0 || val == 1)
			dev->a_stereo = val;
		/*  Supporting only stereo for now...  */
		dev->a_stereo = 1;
		return put_user(dev->a_stereo, (int *)arg);
	}
	case SNDCTL_DSP_SPEED:
	{
		get_user_ret(val, (int *)arg, -EFAULT);
		if (val != dev->a_speed)
		{
			audio_set_speed(dev, val);
		}
		return put_user(dev->a_speed, (int *)arg);
	}

	case SNDCTL_DSP_SYNC:
	{
		if (dev->a_open_mode & FMODE_WRITE)
			audio_drain_play_buffer(dev);
		vdma_stop(dev);
		return 0;
	}
	case SNDCTL_DSP_RESET:
	{
		vdma_stop(dev);
		return 0;
	}
	case SNDCTL_DSP_POST:
	{
		vdma_run(dev);
		return 0;
	}
	case SNDCTL_DSP_GETTRIGGER:
	{
		return -EINVAL;/* BUG: */
	}
	case SNDCTL_DSP_SETTRIGGER:
	{
		return -EINVAL;/* BUG: */
	}

	case SNDCTL_DSP_GETISPACE:
	case SNDCTL_DSP_GETOSPACE:
	{
		audio_buf_info	abinfo;

		audio_fragment_size(dev);
		if (dev->a_fragment == 0)
			dev->a_fragment = 4096;
		abinfo.fragstotal = AUDIO_DMA_BUFFER_SIZE / dev->a_fragment;
		abinfo.fragsize = dev->a_fragment;
		if (cmd == SNDCTL_DSP_GETISPACE)
		{
			abinfo.fragments = dev->a_record_data_avail
				/ dev->a_fragment;
			abinfo.bytes = dev->a_record_data_avail;
		}
		else
		{
			abinfo.fragments = dev->a_play_space_avail
				/ dev->a_fragment;
			abinfo.bytes = dev->a_play_space_avail;
		}
		return copy_to_user((void *)arg, &abinfo, sizeof(abinfo))
			? -EFAULT : 0;
	}
	case SNDCTL_DSP_GETBLKSIZE:
	{	
		audio_fragment_size(dev);
		return put_user(dev->a_fragment, (int *)arg);
	}
	case SNDCTL_DSP_SETFRAGMENT:
	{
		return 0;/* BUG: */
	}
#if 0
	case SNDCTL_DSP_GETIPTR:
	case SNDCTL_DSP_GETOPTR:
		/*  These are optional and sometimes not portable, so skip */
		return -EINVAL;
#endif
	}
debug_msg("Unhandled audio ioctl %08X\n",cmd);
	return -ENOIOCTLCMD;
}

static int
audio_mmap(struct file *file, struct vm_area_struct *vma)
{
	return -EINVAL;
}

static loff_t
audio_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static struct file_operations audio_fops =
{
	audio_llseek,
	audio_read,
	audio_write,
	NULL,  /* readdir */
	audio_poll,
	audio_ioctl,
	audio_mmap,
	audio_open,
	NULL,	/* flush */
	audio_close,
	NULL,  /* fsync */
	NULL,  /* fasync */
	NULL,  /* check_media_change */
	NULL,  /* revalidate */
	NULL,  /* lock */
};


/*
 *	D R I V E R   R E G I S T R A T I O N
 */

static void
config_videum_audio(struct videum_device *dev)
{
	if (dev->wavi_type == WINNOV_WAVI_93 ||
	    dev->wavi_type == WINNOV_WAVI_95)
	{
		info_msg("%s-- Not a WAVI-97 or later chip, no audio.\n",
			 dev->shortname);
		return;/* Can't read back audio pointer */
	}
	if (!wavi_initialize_audio(dev))
		return 0;
	if (dev->sram_audio_size == 0)
	{
		err_msg("%s-- wnv.o did not allocate a buffer in SRAM\n",
			dev->shortname);
		return;/* No hardware audio buffer */
	}
	if (!vdma_setup(dev))
	{
		err_msg("Couldn't allocate audio DMA buffers\n");
		return;
	}
	dev->a_format = AFMT_S16_LE;
	dev->a_stereo = 1;
	audio_set_speed(dev, 44100);
	audio_fragment_size(dev);

	dev->dsp_unit = register_sound_dsp(&audio_fops, -1);
	if (dev->dsp_unit < 0)
	{
		err_msg("Couldn't register the sound dsp driver.\n");
		return;
	}
	info_msg("Registered '%s' as OSS dsp device %d\n", 
		 dev->v.name, dev->dsp_unit);
	dev->dsp_registered = 1;
}

static void
config_videum_mixer(struct videum_device *dev)
{
	if (dev->eeprom.bMinorType == MINORTYPE_V2WINNOV || /* Videum2 */
	    dev->eeprom.bMinorType == MINORTYPE_AV)   /* Videum10 AV */
	{
		dev->mixer_unit = register_sound_mixer(&mixer_fops, -1);
		if (dev->mixer_unit < 0)
		{
			err_msg("Couldn't register the sound mixer driver.\n");
			return 0;
		}
		info_msg("Registered '%s' as OSS mixer device %d\n", 
			 dev->v.name, dev->mixer_unit);
		dev->mixer_registered = 1;
		mixer_pciav_init(dev);
	}
}

static void
unconfig_videum(struct videum_device *dev)
{
	vdma_shutdown(dev);
	mixer_shutdown(dev);
	if (dev->dsp_registered)
		unregister_sound_dsp(dev->dsp_unit);
	dev->dsp_registered = 0;
	if (dev->mixer_registered)
		unregister_sound_mixer(dev->mixer_unit);
	dev->mixer_registered = 0;
}

static int
config_videum(struct videum_device *dev)
{
	/*  Unregistered state is -1 because zero is a valid device number */
	dev->mixer_unit = -1;
	dev->dsp_unit = -1;

	if (dev->type == VIDEUM_TYPE_BOGUS)
		return 0;/* No hardware for this device structure  */

	init_waitqueue(&dev->a_play_wait);
	init_waitqueue(&dev->a_record_wait);

	/*  Audio play/record device  */

	config_videum_audio(dev);

	/*  Mixer device  */

	if (dev->eeprom.bMinorType == MINORTYPE_V2WINNOV || /* Videum2 */
	    dev->eeprom.bMinorType == MINORTYPE_AV)   /* Videum10 AV */
	{
		dev->mixer_unit = register_sound_mixer(&mixer_fops, -1);
		if (dev->mixer_unit < 0)
		{
			err_msg("Couldn't register the sound mixer driver.\n");
			return 0;
		}
		info_msg("Registered '%s' as OSS mixer device %d\n", 
			 dev->v.name, dev->mixer_unit);
		dev->mixer_registered = 1;
		mixer_pciav_init(dev);
	}

	return (dev->dsp_registered ||
		dev->mixer_registered);
}


/*
 *	M O D U L E   I N I T   A N D   C L E A N U P
 */

int
init_module(void)
{
	int	i, n = 0;

	for (i = 0; i < WNV_NBOARDS; ++i)
		if (config_videum(&videum[i]))
			++n;
	if (n == 0)
	{
		err_msg("No Videums with audio capability.\n");
		return -ENODEV;/* cleanup will not be called */
	}
	return 0;
}

void
cleanup_module(void)
{
	int	i;

	for (i = 0; i < WNV_NBOARDS; ++i)
		unconfig_videum(&videum[i]);
}
